"""
Version and license information.
"""

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '3.3.0'
__versiondate__ = '2026-03-27'
__license__ = f'Starsim {__version__} ({__versiondate__}) — © 2023-2026 by IDM'
