/*-- scss:rules --*/

.navbar-logo {
    content: url("assets/starsim-logo-dark.png");
}

.navbar {
    background-color: rgb(52, 52, 52);
}

/*-- From https://github.com/quarto-dev/quarto-web/blob/main/theme-dark.scss --*/

/*-- scss:defaults --*/

$blue: #447099;
$red: #D44000;
$orange: #EE6331;
$yellow: #E7B10A;
$green: #72994E;
$teal: #419599;
$cyan: #75A8D9;
$black: #000;
$white: #fff;
$gray: #404041;
$gray-dark: #343a40;
$gray-100: #f8f8f8;
$gray-200: #C2C2C4;
$gray-300: #949494;
$gray-400: #707073;
$gray-500: #495057;
$gray-600: #404041;
$gray-700: #333333;
$gray-800: #242426;
$gray-900: #151515;
$default: $gray-900;
$primary: $cyan;
$secondary: $gray-400;
$success: $green;
$info: $yellow;
$warning: $orange;
$danger: $red;



// Base document colors
$body-bg: $gray-900;
$body-color: $gray-100;
$link-color: $cyan;

$light: rgb(32.4,32.4,34.2);

// Navigation element colors
/*$footer-bg: shade-color($gray-800, 30%);*/
/*$navbar-bg: shade-color($gray-800, 30%);*/
$footer-bg: $gray-900;
$navbar-bg: shade-color($gray-800, 10%);
 

// Code blocks
$code-block-bg-alpha: -.9;

// Bootstrap popovers
$popover-bg: #242424;

// Bootstrap inputs
$input-bg: #242424;
$dropdown-bg: $body-bg;
$dropdown-color: $body-color;
$dropdown-link-color: $body-color;
$dropdown-link-hover-bg: $gray-600;
$dropdown-link-hover-color: #eee;
$nav-tabs-link-hover-border-color: rgb(80, 80, 80);

$border-color: rgba($gray-800, 0.5);

$mono-background-color: rgba(0,0,0,0);
$code-color: rgb(192, 128, 216);
$nav-tabs-link-active-color: $body-color;
$toc-inactive-border: rgba($gray-800, 0.5);
$input-text-background-color: $gray-800;
$input-group-addon-bg: $gray-800;
$carousel-dark-caption-color: $body-color;
$pagination-disabled-bg: rgb(24,24,24);

$text-muted: lighten(#6a737b, 20%);

/*-- scss:rules --*/

/*Homepage custom css overrides */

.bg-color {
  background-color: $body-bg;
}

.hello-output {
  border: 1px solid $gray-700;
}


body.quarto-light .alt-background {
  background-color: #17212B;
  border-top: 1px solid #dee2e6;
  border-bottom: 1px solid #dee2e6;
}

.code-with-filename .code-with-filename-file pre {
  color: #ddd;
}

pre code {
  color: $body-color;
}

figcaption code,
dl code,
h1 code,
h2 code,
h3 code,
h4 code,
h5 code,
h6 code {
  background-color: $mono-background-color;
}

details > summary {
  color: $text-muted;
}